from dataclasses import dataclass
import time

import serial

from voyage_touch.sensor import SensorType

FSR_MAX_READING = 1000
PIEZO_MAX_READING = 160

@dataclass
class SensorReading:
    sensor_type: SensorType
    sensor_id: int
    value: float
    time: int   # nanoseconds since UNIX epic

class TouchSensor:
    """
    Class for communicating via touch sensors. 
    """
    def __init__(self, port, baudrate=115200, timeout=5, num_fsrs=3, num_pzs=3):
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self.num_fsrs = num_fsrs
        self.num_pzs = num_pzs

        self.should_stop = False

        self.arduino = None

    def connect(self, callback) -> bool:
        self.callback = callback

        # Initialse serial connection
        self.arduino = serial.Serial(port='/dev/ttyACM0', baudrate=115200, timeout=.1) 
        print("initialising serial connection...")
        time.sleep(0.5)  # Wait for the serial connection to initialize

        for _ in range(100):
            line = self.arduino.readline()
            try:
                line = line.decode('utf-8').strip()
                tokens = line.split(',')
                if len(tokens) != 3:
                    continue
                
                # initialised successfully
                return True
            except:
                continue

        return False
    
    def close(self):
        self.should_stop = True

    def run(self):
        assert self.arduino is not None

        while True:
            if self.should_stop:
                return

            if self.arduino.in_waiting > 0:
                try:
                    line = self.arduino.readline().decode('utf-8').strip()
                except UnicodeDecodeError:
                    continue
                tokens = line.split(',')
                if len(tokens) != 3:
                    print(f"invalid format received via serial: {line}\n")
                    continue
                try:
                    if tokens[0] == "FSR":
                        sensor_id = int(tokens[1])
                        if sensor_id >= self.num_fsrs:
                            # more sensors in the hardware device than our software supports
                            # therefore silently ignore them
                            continue

                        # Convert FSR value to an integer
                        pz_value = int(tokens[2])
                        # Clamp the value between 0 and FSR_MAX_READING
                        pz_value = max(0, min(FSR_MAX_READING, pz_value)) / FSR_MAX_READING
                        
                        # Broadcast
                        reading = SensorReading(SensorType.FSR, sensor_id, pz_value, time.time_ns())
                        self.callback(reading)

                    elif tokens[0] == "PZ":
                        sensor_id = int(tokens[1])
                        if sensor_id >= self.num_pzs:
                            # more sensors in the hardware device than our software supports
                            # therefore silently ignore them
                            continue

                        # Convert FSR value to an integer
                        pz_value = int(tokens[2])
                        # Clamp the value between 0 and PIEZO_MAX_READING
                        pz_value = max(0, min(PIEZO_MAX_READING, pz_value)) / PIEZO_MAX_READING

                        # Broadcast
                        reading = SensorReading(SensorType.PIEZO, sensor_id, pz_value, time.time_ns())
                        self.callback(reading)
                    else:
                        print(f"expected sensor type to be FSR or PZ, got {tokens[0]}\n")
                        continue

                except ValueError:
                    pass  # Ignore invalid values


        





from collections import deque
from typing import List

import numpy as np

from voyage_touch.assc import ASSCModel, ASSCModelConfig

class FSRHeuristicPolicy(ASSCModel):
    """
    A heuristic policy for sensors that only contain force sensing resistors.
    Each force sensing resistor is a sensor point.
    Predicts touch events, where voltage > 0 is considered true. 
    """
    def __init__(self, num_fsrs: int, floor = 0.02):
        self.config = ASSCModelConfig(
            num_points=num_fsrs,
            num_fsr_sensors=num_fsrs,
            num_piezo_sensors=0,
        )
        self.floor = floor
    
    def get_config(self) -> ASSCModelConfig:
        return self.config
    
    def predict(self, readings: List[deque]) -> np.array:
        res = np.zeros(self.config.num_points)
        for i, reading_queue in enumerate(readings):
            if len(reading_queue) > 0:
                latest = reading_queue.pop()
                if latest[1].value > self.floor:
                    res[i] = 1
        return res



import asyncio
from queue import Queue
from threading import Thread, Lock
from typing import Optional

from voyage_touch.sensor import SensorReading, SensorType, TouchSensor

FSR_MAX_READING = 1000
PIEZO_MAX_READING = 160

class TouchSensorReader:
    """
    A convenient class for reading raw data from touch sensors via serial connections.
    """
    def __init__(self, sensor: TouchSensor):
        self.sensor = sensor
        self.queues = {}
        self.mutex = Lock()

        self.sensor_readings = [None]*(sensor.num_fsrs + sensor.num_pzs)

    
    def register_listener(self, listener_id: int, q: Queue):
        """
        Registers a new listener queue. All sensor data will be written here.
        """
        with self.mutex:
            self.queues[listener_id] = q

    def deregister_listener(self, listener_id: int):
        """
        Deregisters a listener queue. No more data will be written to the queue.
        """
        with self.mutex:
            self.queues.pop(listener_id, None)

    async def start(self) -> Thread:
        success = self.sensor.connect(self.sensor_callback)
        if not success:
            raise ValueError("failed to connect via serial")
        
        print("serial connection initialised!")
        
        sensor_thread = Thread(target=lambda: asyncio.run(self.sensor.run()))
        sensor_thread.start()
        return sensor_thread
    
    def stop(self):
        self.sensor.close()
    
    def sensor_callback(self, reading: SensorReading):
        sensor_id = reading.sensor_id
        if reading.sensor_type is SensorType.PIEZO:
            sensor_id += self.sensor.num_fsrs

        with self.mutex:
            for q in self.queues.values():
                q.put(reading)
        self.sensor_readings[sensor_id] = reading

    def read(self, sensor_type: SensorType, sensor_id: int) -> Optional[SensorReading]:
        """
        Returns the current sensor reading for the current sensor ID.
        """
        if sensor_type is SensorType.PIEZO:
            sensor_id += self.sensor.num_fsrs

        if sensor_id >= len(self.sensor_readings) or sensor_id < 0:
            raise ValueError("sensor id outside expected range - no such sensor exists")
        
        return self.sensor_readings[sensor_id]



import asyncio
from collections import deque
import time
from threading import Thread

import numpy as np

from voyage_touch.assc import ASSCModel
from voyage_touch.sensor import SensorType, SensorReading, TouchSensor

class ASSC:
    """
    Artificial somatosensory cortex.

    Predicts tactile events based on sensor data.
    """
    def __init__(self, model: ASSCModel, sensor: TouchSensor, sensor_reading_time_window=5000):
        """
        sensor_reading_time_window: time window in milliseconds.
        """
        self.sensor = sensor
        self.model = model
        config = model.get_config()

        assert sensor.num_fsrs == config.num_fsr_sensors
        assert sensor.num_pzs == config.num_piezo_sensors
        
        self.time_window = sensor_reading_time_window * 1_000_000   # convert to nanoseconds
        self.num_points = config.num_points
        self.num_fsr_sensors = config.num_fsr_sensors
        self.num_piezo_sensors = config.num_piezo_sensors

        # maintain a time window of sensor readings for each sensor
        self.readings = [deque()]*(sensor.num_fsrs + sensor.num_pzs)
    
    async def start(self) -> Thread:
        success = self.sensor.connect(self.sensor_callback)
        if not success:
            raise ValueError("failed to connect via serial")
        
        print("serial connection initialised!")
        
        sensor_thread = Thread(target=lambda: asyncio.run(self.sensor.run()))
        sensor_thread.start()
        return sensor_thread
    
    def stop(self):
        self.sensor.close()

    def sensor_callback(self, reading: SensorReading):
        sensor_id = reading.sensor_id
        if reading.sensor_type is SensorType.PIEZO:
            sensor_id += self.sensor.num_fsrs

        current_time = time.time_ns()
        self.readings[sensor_id].append((current_time, reading))
        self._remove_old_readings(sensor_id, current_time)

    def _remove_old_readings(self, sensor_id, current_time):
        while self.readings[sensor_id] and (current_time - self.readings[sensor_id][0][0]) > self.time_window:
            self.readings[sensor_id].popleft()


    def predict(self) -> np.array:
        """
        For every point in the ASSC model, predict the current touch event state.
        """
        return self.model.predict(self.readings)



from abc import ABC, abstractmethod
from collections import deque
from dataclasses import dataclass
from typing import List

import numpy as np

@dataclass
class ASSCModelConfig:
    """
    Config for an ASSC model.
    """
    num_points: int
    num_fsr_sensors: int
    num_piezo_sensors: int

class ASSCModel(ABC):
    """
    An abstract class for an ASSC Model. Should be implemented by all ASSC models.
    """

    @abstractmethod
    def get_config(self) -> ASSCModelConfig:
        """
        Retuns the ASSC config of the model. Used by the ASSC class for managing sensor readings.
        """
        pass

    def predict(self, readings: List[deque]) -> np.array:
        """
        Predict touch events (1 or 0) for each point when given a time series of sensor readings for each sensor.
        """
        pass
